import hashlib
import os
import zipfile
from glob import glob
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import gdown
import numpy as np
import yaml
from numpy.typing import NDArray
from PIL import Image
from roifile import roiread
from scipy.ndimage import gaussian_filter

current_script_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.dirname(current_script_path)
CONFIG_PATH = os.path.join(root_path, "config", "config.yaml")

with open(CONFIG_PATH, "r") as file:
    config = yaml.load(file, Loader=yaml.FullLoader)


def remove_img_without_roi(location: str, remove: bool = True) -> None:
    """Find all images without roi data, write txt file with paths to problems.

    Args:
        location (str): dir with slides. In this location must be dir slides
            and inside must be all slides.
        remove (bool, optional): if True - remove all files without roi data.
            Defaults to True.
    """
    img_location = os.path.join(location, "slides")
    image_list = glob(os.path.join(img_location, "*.tif"))

    error_list = []

    for roi_path in image_list:
        try:
            roiread(roi_path)[1]
        except Exception as e:
            print(f'Warning: "{e}". File {roi_path} was deleted')
            if remove:
                os.remove(roi_path)
            error_list.append(roi_path)
            continue
    if error_list:
        if isinstance(config["errors_path"], List):
            errors_path = "/".join(config["errors_path"])
        elif isinstance(config["errors_path"], str):
            errors_path = config["errors_path"]

        with open(errors_path, "w") as file:
            for path in error_list:
                file.write(path + "\n")


def get_and_unzip(url: str, location: str) -> None:
    """Extract a ZIP archive from given URL.

    Args:
        url (str): google drive url id of a ZIP file
        location (str): target location to extract archive in
    """
    dataset = gdown.download(id=url, fuzzy=True, output="polony.zip")
    dataset = zipfile.ZipFile(dataset)
    dataset.extractall(location)
    dataset.close()
    os.remove(dataset.filename)
    remove_img_without_roi(location)


def read_tiff(path: str, new_size: Optional[Tuple[int]] = None) -> np.ndarray:
    """Read tiff file from path, make resize and return numpy array wth image.

    Args:
        path (str): Path to the multipage-tiff file.
        new_size (Optional[Tuple], optional): If not None - resize img
            to new_size. Defaults to None.

    Returns:
        np.ndarray: Return numpy array with image
    """
    img = Image.open(path)
    images = []
    for i in range(img.n_frames):
        img.seek(i)
        if new_size is not None:
            images.append(np.array(img.resize((new_size[1], new_size[0]))))
        else:
            images.append(np.array(img))
    return np.array(images, dtype=np.float32) / 255


def get_roi_coordinates(
    roi_path: str, channel: Union[int, None] = None, counter: bool = False
) -> Union[NDArray[np.float64], Tuple[NDArray[np.float64]]]:
    """
    From ROI file with image get arrays with coordinates
    Args:
        roi_path (str): path to ROI image
        channel (Union[int, None], optional): for which channel of the image
        you need to return an array with coordinates.
            Defaults to None - for both. Can be 1 or 2
        counter (False): return or not also array with points areas id

    Returns:
        NDArray[np.float64] or Tuple[NDArray[np.float64]]]:
            Numpy array with points coordinates
            or if counter == True
                tuple with 2 arrays: (point coordinates, point area id)
            or if channel is None
                point coordinates - tuple with two arrays
                point area id - tuple with two arrays
    """
    try:
        img_roi = roiread(roi_path)[1]
    except ValueError as e:
        print(f'ERROR "{e}" in file {roi_path}')
        # os.remove(roi_path)
        raise e
    counter_positions = img_roi.counter_positions
    subpixel_coordinates = img_roi.subpixel_coordinates

    if not counter:
        if channel is None:
            coordinates_1 = subpixel_coordinates[counter_positions == 1]
            coordinates_2 = subpixel_coordinates[counter_positions == 2]
            return (coordinates_1, coordinates_2)
        elif channel == 1 or channel == 2:
            return subpixel_coordinates[counter_positions == channel]
        else:
            raise ValueError("Invalid value for variable. Must be 1 or 2")
    counters = img_roi.counters
    if channel is None:
        coordinates_1 = subpixel_coordinates[counter_positions == 1]
        coordinates_2 = subpixel_coordinates[counter_positions == 2]
        return (
            (
                coordinates_1,
                coordinates_2,
            ),
            (
                counters[counter_positions == 1],
                counters[counter_positions == 2],
            ),
        )
    if channel in (1, 2):  # skipcq: PYL-R1714
        return (
            subpixel_coordinates[counter_positions == channel],
            counters[counter_positions == channel],
        )
    else:
        raise ValueError("Invalid value for variable. Must be 1 or 2")


def create_density_roi(
    coordinates: np.ndarray,
    size: Tuple[int] = config["img_size"],
    new_size: Optional[Tuple[int]] = None,
) -> NDArray[np.float64]:
    """Create density map by points coordinates

    Args:
        coordinates (np.ndarray): Points coordinates
        size (Tuple[int], optional): Size of image.
            Defaults to config['img_size'].
        new_size (Optional[Tuple[int]], optional): Size for resize.
            Defaults to None.

    Returns:
        NDArray[np.float64]: Output density map
    """
    if new_size is not None:
        density = np.zeros(new_size, dtype=np.float32)

        # calculate scale coeffs
        scale_y = new_size[0] / size[0]
        scale_x = new_size[1] / size[1]

    else:
        density = np.zeros(size, dtype=np.float32)

    # make a one-channel label array with 100 in dots positions
    for x, y in coordinates:
        if x < size[1] and y < size[0]:
            if new_size is not None:
                # pow to scales coeffs
                x = int(x * scale_x)
                y = int(y * scale_y)
            if x < 0 or y < 0:
                print(x, y)
                return False
            density[int(y), int(x)] = 100

    # generate a density map by applying a Gaussian filter
    density = gaussian_filter(density, sigma=(1, 1), order=0)

    return density


def reshape_numpy(arr: np.ndarray, new_size: List[int]) -> np.ndarray:
    """Reshape numpy arrays (3d and 2d)

    Args:
        arr (np.ndarray): source vector
        new_size (List[int]): new size to reshape

    Returns:
        np.ndarray: numpy array after reshape to new_size
    """
    dim = len(arr.shape)
    if dim not in [2, 3]:
        raise ValueError("Shape mismatch")
    h, w = new_size[::-1]
    if dim == 3:
        new_arr = np.transpose(arr, (1, 2, 0))
        new_arr = cv2.resize(new_arr, (w, h))
        return np.transpose(new_arr, (2, 0, 1))
    return cv2.resize(arr, (w, h))


def grid_to_squares(
    path: str,
    new_size: Optional[List[int]] = None,
    mode: str = "density",
) -> List[Dict[str, Any]]:
    """Make dict with squares and other data for model training
        from original tiff file with image

    Args:
        path (str): path to tiff file with image
        new_size (List[int] | None): new_size for squares
        mode (str): for which model should be prepared squares 'density' or 'classifier'

    Returns:
        List[Dict[str, Any]]: List with dicts. Each dict for square image and
            training data for it
    """
    # Uploading an image
    imgs = read_tiff(path)
    img = imgs[0]

    img_roi = roiread(path)

    points, counters = get_roi_coordinates(
        roi_path=path,
        # TODO Think about how to make it so that it is automatically
        # determined on which channels there are points
        channel=1,
        counter=True,
    )

    lines_coordinates = img_roi[0].multi_coordinates.reshape(-1, 6).astype(int)[:, 1:3]

    # square size
    square_size = lines_coordinates[1, 0] - lines_coordinates[0, 0]

    # Dividing coordinates into horizontal and vertical lines
    horizontal_lines = lines_coordinates[lines_coordinates[:, 1] == 0][:, 0]
    vertical_lines = lines_coordinates[lines_coordinates[:, 0] == 0][:, 1]

    squares = []  # List for storing full squares

    # Splitting an image into squares
    for x in horizontal_lines[:-1]:
        for y in vertical_lines[:-1]:
            points_condition = (
                (points[:, 1] >= y)
                & (points[:, 1] < y + square_size)
                & (points[:, 0] >= x)
                & (points[:, 0] < x + square_size)
            )

            square_points = points[points_condition]

            square_id = counters[points_condition]
            id_value, id_counts = np.unique(square_id, return_counts=True)

            # skip squares with 1 or 0 points or with more than 2 types of points
            if len(square_points) == 1 or (
                len(id_value) > 1 and max(id_counts) - min(id_counts) <= 2
            ):
                continue
            elif len(square_points) == 0:
                square_class = 0
                # skip empty squares for 'density' mode
                true_square_id = -1
                if mode == "density":
                    continue
            else:
                square_class = 1
                true_square_id = id_value[id_counts == max(id_counts)].item()

            # CHECK AND BRING BACK TO SQUARE POINTS if points are out of bounds
            # ###TODO: (28.11) may be better idea to delete such squares
            square_points = bring_back_points(
                true_square_id, points, counters, x, y, square_size
            )

            square_dict = dict()
            square = img[y : y + square_size, x : x + square_size]
            square_2c = imgs[:, y : y + square_size, x : x + square_size]

            if new_size is None:
                square_dict["square"] = square
                square_dict["square_2c"] = square_2c
            else:
                square = np.transpose(square)
                square_dict["square"] = reshape_numpy(square, new_size)
                square_dict["square_2c"] = reshape_numpy(square_2c, new_size)
            # points in relative coordinates for a given square
            square_dict["points"] = square_points
            # coordinates of the upper-left corner of the square
            square_dict["square_coordinate"] = np.array([x, y])
            square_dict["n_points"] = len(square_points)

            square_dict["label"] = create_density_roi(
                square_dict["points"], size=[square_size] * 2, new_size=new_size
            )
            square_dict["class"] = square_class
            if square_dict["label"] is False and square_class == 1:
                print(
                    "x ",
                    x,
                    "y ",
                    y,
                    path,
                    "id ",
                    true_square_id,
                    "id_value ",
                    id_value,
                )
            squares.append(square_dict)

    return squares


def bring_back_points(
    square_id: int,
    points: NDArray[np.float64],
    counters: NDArray[np.float64],
    x: int,
    y: int,
    square_size: int,
) -> NDArray[np.float64]:
    """This function makes INPLACE changes in the list points

    Args:
        square_id (int): id of square
        points (NDArray[np.float64]): list with coordinates (full list)
        counters (NDArray[np.float64]): list with points id (same id -> same square)
        x (int): x coordinate of square
        y (int): y coordinate of square
        square_size (int): size of square

    Returns:
        NDArray[np.float64]: list of correct points of square with square_id
    """
    if square_id == -1:
        return np.array([])
    true_points = points[counters == square_id]  # square points by id of square
    points_condition = (
        (true_points[:, 1] >= y)
        & (true_points[:, 1] < y + square_size)
        & (true_points[:, 0] >= x)
        & (true_points[:, 0] < x + square_size)
    )

    points_anti_condition = np.invert(
        points_condition
    )  # bool list with true on points which is not in square

    if (
        np.sum(points_anti_condition) == 0
    ):  # check if we don't have any points out of square
        return true_points - np.array([x, y])  # return list of square_points
    mistake_points = true_points[points_anti_condition]
    for i, (p_x, p_y) in enumerate(mistake_points):
        if p_x < x:
            mistake_points[i][0] = x
        if p_x >= x + square_size:
            mistake_points[i][0] = x + square_size - 0.5
        if p_y < y:
            mistake_points[i][1] = y
        if p_y >= y + square_size:
            mistake_points[i][1] = y + square_size - 0.5
    true_points[points_anti_condition] = mistake_points
    return true_points - np.array([x, y])


def count_data_size(image_list, mode="density"):
    counter = 0
    for img_path in image_list:
        squares_list = grid_to_squares(img_path, mode=mode)
        counter += len(squares_list)

    return counter


def rgb_to_gray(my_data: np.ndarray):
    example = np.transpose(my_data, (1, 2, 0))
    example = cv2.cvtColor(example, cv2.COLOR_RGB2GRAY)
    example = np.expand_dims(example, axis=0)
    return example


def mean_std(data_train):
    mean = np.array([0.0, 0.0])
    std = np.array([0.0, 0.0])
    for i in range(len(data_train)):
        mean += data_train[i][0].mean((1, 2))
        std += data_train[i][0].std((1, 2))
    return mean / len(data_train), std / len(data_train)


def delete_duplicates(image_folder):
    # Creating a dictionary where the key is the hash of the image,
    # and the value is a list of files with this hash
    image_hash_dict = {}

    # Going through all the files in the folder
    for root, dirs, files in os.walk(image_folder):
        if len(files) == 0:
            continue
        for filename in files:
            file_path = os.path.join(root, filename)

            # Skip the non-images
            try:
                with Image.open(file_path) as img:
                    image_hash = hashlib.md5(img.tobytes()).hexdigest()
            except Exception as e:
                print(f"Error during file processing {file_path}: {e}")
                continue

            # If the hash of the image is already in the dictionary,
            # add the file to the corresponding list
            if image_hash in image_hash_dict:
                image_hash_dict[image_hash].append(file_path)
            else:
                # If there is no hash, create a new entry in the dictionary
                image_hash_dict[image_hash] = [file_path]

    # Deleting duplicate files
    for image_hash, file_list in image_hash_dict.items():
        if len(file_list) > 1:
            print(f"Removing duplicates for the hash {image_hash}:")
            for file_path in file_list[1:]:
                os.remove(file_path)
                print(f"  Deleted file: {file_path}")

    print("Deletion completed.")



import os
from typing import Dict, List, Tuple

import numpy as np
import torch
import wandb
from numpy.typing import NDArray
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)
from torchvision import transforms

from ..data.utils import grid_to_squares, read_tiff
from .models import UNet

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
FIRST_HORIZONTAL = 158
FIRST_VERTICAL = 158
SQUARE_SIZE = [316, 316]
# [vertical lines, horizontal lines] = [8, 5] by ImageJ
NUMBER_OF_LINES = [8, 5]

MEAN, STD = ([12.69365111, 2.47628206], [13.35308926, 2.45260453])
normalize = transforms.Normalize(MEAN, STD)


class Looper:
    """Looper handles epoch loops, logging, and plotting."""

    def __init__(
        self,
        network: torch.nn.Module,
        device: torch.device,
        loss: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        data_loader: torch.utils.data.DataLoader,
        dataset_size: int,
        validation: bool = False,
        regressor=None,
        batch_idx=0,
        relative_error=False,
        wandb_bool=False,
        transforms=None,
        mode: str = "density",
    ):
        """
        Initialize Looper.

        Args:
            network: already initialized model
            device: a device model is working on
            loss: the cost function
            optimizer: already initialized optimizer link to network parameters
            data_loader: already initialized data loader
            dataset_size: no. of samples in dataset
            plot: matplotlib axes
            validation: flag to set train or eval mode
            regressor: None or model for counting objects from network output
            mode (str): for which model Looper works 'density' or 'classifier'

        """
        self.network = network
        self.device = device
        self.loss = loss
        self.optimizer = optimizer
        self.loader = data_loader
        self.size = dataset_size
        self.validation = validation
        self.running_loss: List[float] = []
        self.regressor = regressor
        self.batch_idx = batch_idx
        self.relative_error = relative_error
        self.wandb_bool = wandb_bool
        self.transforms = transforms
        self.mode = mode

    def run(self) -> float | Tuple[float]:
        """Run a single epoch loop.

        Returns:
            Mean absolute error.
        """
        if self.mode == "classifier":
            return self._run_classifier()

        # reset current results and add next entry for running loss
        self.true_values = []
        self.predicted_values = []
        self.running_loss.append(0)

        if self.regressor is None:
            # set a proper mode: train or eval
            self.network.train(not self.validation)
        else:
            self.regressor.train(not self.validation)
            self.network.train(False)

        for i, (image, label, n_points, path) in enumerate(self.loader):
            # move images and labels to given device
            image = self.transforms(image).to(self.device)
            label = label.to(self.device)
            n_points = n_points.to(self.device)

            # clear accumulated gradient if in train mode
            if not self.validation:
                self.optimizer.zero_grad()

            # get model prediction (a density map)
            result = self.network(image)

            if self.regressor is not None:
                result = self.regressor(result)

            # calculate loss and update running loss
            if self.regressor is None:
                loss = self.loss(result, label)
            else:
                loss = self.loss(result, n_points)
            self.running_loss[-1] += image.shape[0] * loss.item() / self.size

            # update weights if in train mode
            if not self.validation:
                loss.backward()
                self.optimizer.step()

            if self.regressor is None:
                # loop over batch samples
                for true, predicted in zip(label, result):
                    # integrate a density map to get no. of objects
                    # note: density maps were normalized to 100 * no.of objects
                    # to make network learn better
                    true_counts = torch.sum(true).item() / 100
                    predicted_counts = torch.sum(predicted).item() / 100

                    # update current epoch results
                    self.true_values.append(true_counts)
                    self.predicted_values.append(predicted_counts)

            else:
                # loop over batch samples
                for true_counts, predicted_counts in zip(n_points, result):
                    # update current epoch results
                    self.true_values.append(true_counts.item())
                    self.predicted_values.append(torch.round(predicted_counts).item())

        # calculate errors and standard deviation
        self.update_errors()

        # print epoch summary
        self.log()
        if self.relative_error:
            return self.mean_abs_rel_err, self.mean_abs_err
        return self.mean_abs_err

    def freeze_layers(self, mode: str) -> None:
        if isinstance(self.network, torch.nn.DataParallel):
            self.network.module.freeze_layers(mode)
        else:
            self.network.freeze_layers(mode)

    def _run_classifier(self) -> float:
        # reset current results and add next entry for running loss
        self.true_values = []
        self.predicted_values = []
        self.running_loss.append(0)

        self.network.train(not self.validation)

        for i, (image, image_class, _) in enumerate(self.loader):
            # move images and images classes to given device
            image = self.transforms(image).to(self.device)
            image_class = image_class.to(self.device)

            # clear accumulated gradient if in train mode
            if not self.validation:
                self.optimizer.zero_grad()

            # get classifier prediction
            result = self.network(image)

            # calculate loss and update running loss
            loss = self.loss(result, image_class)

            self.running_loss[-1] += image.shape[0] * loss.item() / self.size

            # update weights if in train mode
            if not self.validation:
                loss.backward()
                self.optimizer.step()

            # loop over batch samples
            for true, predicted in zip(image_class, result):
                # update current epoch results
                self.true_values.append(true.item())
                self.predicted_values.append(predicted.item())

        # calculate errors and standard deviation
        if self.mode == "classifier":
            self.predicted_values = self.to_binary(np.array(self.predicted_values))
        self.update_errors()

        # print epoch summary
        self.log()
        if self.mode == "classifier":
            return self.f1
        if self.relative_error:
            return self.mean_abs_rel_err, self.mean_abs_err
        return self.mean_abs_err

    def sigmoid(self, x: NDArray) -> NDArray:
        return 1 / (1 + np.exp(-x))

    def to_binary(self, logits: NDArray, threshold=0.5) -> List[float]:
        probabilities = self.sigmoid(logits)
        return (probabilities > threshold).astype(float).tolist()

    def update_errors(self):
        """
        Calculate errors and standard deviation based on current
        true and predicted values.
        """
        if self.mode == "classifier":
            self.accuracy = accuracy_score(self.true_values, self.predicted_values)
            self.precision = precision_score(self.true_values, self.predicted_values)
            self.recall = recall_score(self.true_values, self.predicted_values)
            self.f1 = f1_score(self.true_values, self.predicted_values)
            self.roc_auc = roc_auc_score(self.true_values, self.predicted_values)
            self.confusion = confusion_matrix(self.true_values, self.predicted_values)
            stage = "train" if not self.validation else "val"
            metrics = {
                f"{stage}/loss": self.running_loss[-1],
                f"{stage}/accuracy": self.accuracy,
                f"{stage}/precision": self.precision,
                f"{stage}/recall": self.recall,
                f"{stage}/f1": self.f1,
                f"{stage}/confusion": self.confusion,
            }

        elif self.mode == "density":
            self.err = [
                true - predicted
                for true, predicted in zip(self.true_values, self.predicted_values)
            ]
            self.relative_err = [
                (true - predicted) / true
                for true, predicted in zip(self.true_values, self.predicted_values)
            ]
            self.square_err = [error * error for error in self.err]
            self.abs_err = [abs(error) for error in self.err]
            self.abs_rel_err = [abs(error) for error in self.relative_err]
            self.mean_err = sum(self.err) / self.size
            self.mean_abs_err = sum(self.abs_err) / self.size
            self.mean_abs_rel_err = sum(self.abs_rel_err) / self.size
            self.mean_square_error = sum(self.square_err) / self.size
            self.std = np.array(self.err).std()

            stage = "train" if not self.validation else "val"
            metrics = {
                f"{stage}/loss": self.running_loss[-1],
                f"{stage}/mean_err": self.mean_err,
                f"{stage}/MAE": self.mean_abs_err,
                f"{stage}/MARE": self.mean_abs_rel_err,
                f"{stage}/std": self.std,
                f"{stage}/MSE": self.mean_square_error,
            }

        if self.wandb_bool:
            wandb.log(metrics)

    def log(self):
        """Print current epoch results."""
        if self.mode == "density":
            print(
                f"{'Train' if not self.validation else 'Valid'}:\n"
                f"\tAverage loss: {self.running_loss[-1]:3.4f}\n"
                f"\tMean error: {self.mean_err:3.3f}\n"
                f"\tMean absolute error: {self.mean_abs_err:3.3f}\n"
                f"\tMean absolute relative error: {self.mean_abs_rel_err:1.4f}\n"
                f"\tError deviation: {self.std:3.3f}\n"
                f"\tMean square error: {self.mean_square_error:3.3f}"
            )
        elif self.mode == "classifier":
            print(
                f"{'Train' if not self.validation else 'Valid'}:\n"
                f"\tAverage loss: {self.running_loss[-1]:3.4f}\n"
                f"\tAccuracy: {self.accuracy:3.3f}\n"
                f"\tPrecision: {self.precision:3.3f}\n"
                f"\tRecall: {self.recall:3.3f}\n"
                f"\tF1: {self.f1:3.3f}\n"
                f"\tConfusion matrix:\n{self.confusion}"
            )


class Config(dict):
    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")


def evaluate(
    path_to_example: str,
    path_to_model: str,
) -> Tuple[float, int]:
    """
    args:
        path_to_example - path to one image
        path_to_model - path to saved state dict of model
    return:
        mean absolute error on squares with points,
        total number of points in image
    """

    network = torch.nn.DataParallel(UNet(res=False))
    network.load_state_dict(torch.load(path_to_model, map_location=device))
    network = network.eval()
    example = grid_to_squares(path_to_example)
    error = 0
    abs_result = 0
    for i in range(len(example)):
        img_1 = example[i]["square_2c"]
        img_1 = torch.from_numpy(img_1).float()
        density = network(img_1.unsqueeze(0))
        n_points = example[i]["n_points"]
        result = torch.sum(density).item() // 100
        error += abs(int(n_points) - int(result))
        abs_result += int(result)
    return error / len(example), abs_result


def predict(
    path: str,
    path_to_model: str,
    model: torch.nn.Module = UNet(res=False),
    device: torch.device = torch.device(
        "cuda:0" if torch.cuda.is_available() else "cpu"
    ),
    channels: int = 2,
) -> List[Dict[int, int]]:
    """
    args:
        path: str - path to folder or to file
        path_to_model - path to saved state dict
        model - network which state dict was saved
    return:
        prediction - list with dictionaries;
                     one dict for one image;
                     dict = {square_id: number of points}
    """
    is_dir = os.path.isdir(path)

    predictions = []

    if is_dir:
        images = os.listdir(path)
        print("Files and folders in the directory:")
        for image_path in images:
            predictions.append(
                predict_one_image(image_path, path_to_model, model, device, channels)
            )
    else:
        predictions.append(
            predict_one_image(path, path_to_model, model, device, channels)
        )

    return predictions


def predict_one_image(
    path: str,
    path_to_model: str,
    model: torch.nn.Module = UNet(res=False),
    device: torch.device = torch.device(
        "cuda:0" if torch.cuda.is_available() else "cpu"
    ),
    channels: int = 2,
) -> Dict[int, int]:
    """
    args:
        path - path to image
        path_to_model - path to saved state dict of model
        model - network which state dict was saved

        channels - 1 or 2 channels of image we need (depends on model)

    return:
        dictionary: {square_id: number_of_points}
    """
    network = torch.nn.DataParallel(model).to(device)
    network.load_state_dict(torch.load(path_to_model, map_location=device))
    network = network.eval()

    if channels not in [1, 2]:
        raise ValueError("Not correct value of channels, must be 1 or 2")

    imgs = read_tiff(path)
    img = imgs[0]

    squares_dict = dict()
    # TODO find mean value of first lines
    first_horizontal = FIRST_HORIZONTAL
    first_vertical = FIRST_VERTICAL

    # square size TODO find constant value of square_size
    square_size = SQUARE_SIZE[0]

    # vertical (8) and horizontal (5) lines
    v, h = NUMBER_OF_LINES
    horizontal_lines = [first_horizontal + square_size * i for i in range(h)]
    vertical_lines = [first_vertical + square_size * i for i in range(v)]

    # Devide image to squares and predict n points
    square_id = -1
    for x in horizontal_lines[:-1]:
        for y in vertical_lines[:-1]:
            result_dict = dict()
            square_id += 1
            if channels == 1:
                square = img[y : y + square_size, x : x + square_size]
            elif channels == 2:
                square = imgs[:, y : y + square_size, x : x + square_size]
            square = torch.from_numpy(square).float().to(device)
            square = normalize(square)
            density = network(square.unsqueeze(0))
            result = torch.sum(density).item() // 100

            result_dict["result"] = int(result)
            result_dict["density"] = density
            squares_dict[square_id] = result_dict

    return squares_dict



from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models import ResNet18_Weights, resnet18


class ConvCat(nn.Module):
    """Convolution with upsampling + concatenate block."""

    def __init__(
        self,
        channels: Tuple[int, int],
        size: Tuple[int, int],
        stride: Tuple[int, int] = (1, 1),
        N: int = 1,
        res: bool = False,
    ):
        """
        Create a sequential container with convolutional block (see conv_block)
        with N convolutional layers and upsampling by factor 2.
        """
        super(ConvCat, self).__init__()
        self.conv = nn.Sequential(
            conv_block(channels, size, stride, N),
        )
        self.res = res
        if self.res:
            self.conv_skip = nn.Conv2d(*channels, kernel_size=1)

    def forward(self, to_conv: torch.Tensor, to_cat: torch.Tensor):
        """Forward pass.

        Args:
            to_conv: input passed to convolutional block and upsampling
            to_cat: input concatenated with the output of a conv block
        """
        if self.res:
            to_upsample = self.conv(to_conv) + self.conv_skip(to_conv)
        else:
            to_upsample = self.conv(to_conv)

        upsample_conv = F.interpolate(to_upsample, size=to_cat.shape[-2:])
        return torch.cat([upsample_conv, to_cat], dim=1)


def conv_block(
    channels: Tuple[int, int],
    size: Tuple[int, int],
    stride: Tuple[int, int] = (1, 1),
    N: int = 1,
):
    """
    Create a block with N convolutional layers with ReLU activation function.
    The first layer is IN x OUT, and all others - OUT x OUT.

    Args:
        channels: (IN, OUT) - no. of input and output channels
        size: kernel size (fixed for all convolution in a block)
        stride: stride (fixed for all convolution in a block)
        N: no. of convolutional layers

    Returns:
        A sequential container of N convolutional layers.
    """

    # a single convolution + batch normalization + ReLU block
    def block(in_channels: int):
        return nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=channels[1],
                kernel_size=size,
                stride=stride,
                bias=False,
                padding=(size[0] // 2, size[1] // 2),
            ),
            nn.BatchNorm2d(num_features=channels[1]),
            nn.ReLU(),
        )

    # create and return a sequential container of convolutional layers
    # input size = channels[0] for first block and channels[1] for all others
    return nn.Sequential(*[block(channels[bool(i)]) for i in range(N)])


class UNet(nn.Module):
    """
    U-Net implementation.

    Ref. O. Ronneberger et al. "U-net: Convolutional networks for biomedical
    image segmentation."
    """

    def __init__(
        self,
        filters: int = 64,
        input_filters: int = 2,
        N: int = 2,
        res: bool = False,
        **kwargs
    ):
        """
        Create U-Net model with:

            * fixed kernel size = (3, 3)
            * fixed max pooling kernel size = (2, 2) and upsampling factor = 2
            * fixed no. of convolutional layers per block = 2 (see conv_block)
            * constant no. of filters for convolutional layers

        Args:
            filters: no. of filters for convolutional layers
            input_filters: no. of input channels
        """
        super(UNet, self).__init__()
        # first block channels size
        initial_filters = (input_filters, filters)
        # channels size for downsampling
        down_filters = (filters, filters)
        # channels size for upsampling (input doubled because of concatenate)
        up_filters = (2 * filters, filters)

        self.res = res

        # downsampling
        self.block1 = conv_block(channels=initial_filters, size=(3, 3), N=N)
        self.block2 = conv_block(channels=down_filters, size=(3, 3), N=N)
        self.block3 = conv_block(channels=down_filters, size=(3, 3), N=N)

        # upsampling
        self.block4 = ConvCat(channels=down_filters, size=(3, 3), N=N, res=res)
        self.block5 = ConvCat(channels=up_filters, size=(3, 3), N=N, res=res)
        self.block6 = ConvCat(channels=up_filters, size=(3, 3), N=N, res=res)

        # density prediction
        self.block7 = conv_block(channels=up_filters, size=(3, 3), N=N)
        self.density_pred = nn.Conv2d(
            in_channels=filters, out_channels=1, kernel_size=(1, 1), bias=False
        )

        # residual
        if self.res:
            self.conv_skip1 = nn.Conv2d(*initial_filters, kernel_size=1)
            self.conv_skip2 = nn.Conv2d(*down_filters, kernel_size=1)
            self.conv_skip3 = nn.Conv2d(*down_filters, kernel_size=1)

    def forward(self, input: torch.Tensor):
        """Forward pass."""
        # use the same max pooling kernel size (2, 2) across the network
        pool = nn.MaxPool2d(2)

        # downsampling
        if self.res:
            block1 = self.block1(input) + self.conv_skip1(input)
        else:
            block1 = self.block1(input)
        pool1 = pool(block1)
        if self.res:
            block2 = self.block2(pool1) + self.conv_skip2(pool1)
        else:
            block2 = self.block2(pool1)
        pool2 = pool(block2)
        if self.res:
            block3 = self.block3(pool2) + self.conv_skip3(pool2)
        else:
            block3 = self.block3(pool2)
        pool3 = pool(block3)

        # upsampling
        block4 = self.block4(pool3, block3)
        block5 = self.block5(block4, block2)
        block6 = self.block6(block5, block1)

        # density prediction
        block7 = self.block7(block6)
        return self.density_pred(block7)


class Classifier(nn.Module):
    def __init__(self, in_channel: int = 2) -> None:
        super().__init__()
        self.in_channel = in_channel
        self.weights = ResNet18_Weights.DEFAULT
        self.net = resnet18(weights=self.weights)
        self.net.fc = nn.Linear(self.net.fc.in_features, 1)
        if self.in_channel != 3:
            self.net.conv1 = nn.Conv2d(
                self.in_channel,
                self.net.conv1.out_channels,
                kernel_size=(7, 7),
                stride=(2, 2),
                padding=(3, 3),
                bias=False,
            )

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        return self.net(inp)

    def freeze_layers(self, mode: str = "on") -> None:
        if mode == "on":
            for param in self.parameters():
                param.requires_grad_ = False
            self.net.fc.requires_grad_ = True
            if self.in_channel != 3:
                self.net.conv1.requires_grad_ = True
        elif mode == "off":
            for param in self.parameters():
                param.requires_grad_ = True
        else:
            raise ValueError("Wrong mode, should be 'on' or 'off'")



import os

import numpy as np
import torch
import wandb
import yaml
from torch import nn
from torchvision import transforms

from ..data.make_dataset import PolonyDataset
from ..data.utils import mean_std
from .models import Classifier, UNet
from .utils import Config, Looper

# folder to load config file
current_script_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.dirname(current_script_path)
CONFIG_PATH = os.path.join(root_path, "config", "config.yaml")

with open(CONFIG_PATH, "r") as file:
    config_yaml = yaml.load(file, Loader=yaml.FullLoader)
train_params = config_yaml["train"]


def train(
    dataset_name: str = train_params["dataset_name"],
    network_architecture: str | nn.Module = train_params["network_architecture"],
    learning_rate: float = train_params["learning_rate"],
    epochs: int = train_params["epochs"],
    batch_size: int = train_params["batch_size"],
    unet_filters: int = train_params["unet_filters"],
    convolutions: int = train_params["convolutions"],
    lr_patience: int = train_params["lr_patience"],
    input_channels: int = train_params["input_channels"],
    wandb_bool: bool = train_params["wandb_bool"],
    factor: float = train_params["factor"],
    res: bool = train_params["res"],
    loss: nn.MSELoss = nn.MSELoss(),
    freeze_threshold: int = 10,
) -> None:
    """Train chosen model on selected dataset.

    Args:
        dataset_name (str, optional): _description_.
            Defaults to train_params["dataset_name"].
        network_architecture (str | nn.Module, optional): _description_.
            Defaults to train_params["network_architecture"].
        learning_rate (float, optional): _description_.
            Defaults to train_params["learning_rate"].
        epochs (int, optional): _description_.
            Defaults to train_params["epochs"].
        batch_size (int, optional): _description_.
            Defaults to train_params["batch_size"].
        unet_filters (int, optional): _description_.
            Defaults to train_params["unet_filters"].
        convolutions (int, optional): _description_.
            Defaults to train_params["convolutions"].
        lr_patience (int, optional): _description_.
            Defaults to train_params["lr_patience"].
        input_channels (int, optional): _description_.
            Defaults to train_params["input_channels"].
        wandb_bool (bool, optional): _description_.
            Defaults to train_params["wandb_bool"].
        factor (float, optional): _description_.
            Defaults to train_params["factor"].
        res (bool, optional): _description_.
            Defaults to train_params["res"].
        loss (nn.MSELoss, optional): _description_.
            Defaults to nn.MSELoss().
        freeze_threshold (int, optional): _description_.
            Defaults to 10.
    """

    # use GPU if avilable
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # update params of training
    if network_architecture == "UNet":
        looper_mode = "density"
        lr_mode = "min"
    elif network_architecture == "Classifier":
        looper_mode = "classifier"
        lr_mode = "max"

    if wandb_bool:
        # start a new wandb run to track this script
        run = wandb.init(
            # set the wandb project where this run will be logged
            project="polony",
            save_code=True,
            # track hyperparameters and run metadata
            config={
                "learning_rate": learning_rate,
                "architecture": network_architecture,
                "dataset": dataset_name,
                "epochs": epochs,
                "lr_patience": lr_patience,
                "factor": factor,
            },
        )

        # Copy config to WandB
        config = wandb.config
    #         artifact = wandb.Artifact(name='neural_network', type='model')
    else:
        config = Config(
            {
                "learning_rate": learning_rate,
                "architecture": network_architecture,
                "dataset": dataset_name,
                "epochs": epochs,
                "lr_patience": lr_patience,
                "factor": factor,
            }
        )

    dataset = {}  # training and validation HDF5-based datasets
    dataloader = {}  # training and validation dataloaders
    shuffle = {"train": True, "valid": False}

    for mode in ["train", "valid"]:
        # expected HDF5 files in dataset_name/(train | valid).h5
        # turn on flips only for training dataset
        polony_dataset_params = (
            config_yaml["PolonyDataset_train"]
            if mode == "train"
            else config_yaml["PolonyDataset_val"]
        )
        dataset[mode] = PolonyDataset(**polony_dataset_params, mode=looper_mode)
        dataloader[mode] = torch.utils.data.DataLoader(
            dataset[mode], batch_size=batch_size, shuffle=shuffle[mode]
        )
    # initialize a model based on chosen network_architecture
    if network_architecture == "UNet":
        network = UNet(
            input_filters=input_channels,
            filters=unet_filters,
            N=convolutions,
            res=res,
        ).to(device)
    elif network_architecture == "Classifier":
        network = Classifier(in_channel=input_channels).to(device)
        loss = nn.BCEWithLogitsLoss(
            pos_weight=torch.tensor(dataset["train"].pos_weight())
        )
    elif isinstance(network_architecture, nn.Module):
        network = network_architecture.to(device)
    else:
        raise ValueError(
            "Wrong network, shoud be nn.Module or string: 'Classifier' or 'UNet'"
        )
    network = torch.nn.DataParallel(network)
    optimizer = torch.optim.AdamW(
        network.parameters(),
        lr=config.learning_rate,
    )

    lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer,
        mode=lr_mode,
        patience=config.lr_patience,
        verbose=True,
        factor=config.factor,
    )

    MEAN, STD = mean_std(dataset["train"])
    normalize = transforms.Normalize(MEAN, STD)
    # create training and validation Loopers to handle a single epoch
    train_looper = Looper(
        network,
        device,
        loss,
        optimizer,
        dataloader["train"],
        len(dataset["train"]),
        wandb_bool=wandb_bool,
        transforms=normalize,
        mode=looper_mode,
    )
    # turn on freezing layers for better learning
    if looper_mode == "classifier":
        train_looper.freeze_layers("on")
        print("Layers were frozen")

    valid_looper = Looper(
        network,
        device,
        loss,
        optimizer,
        dataloader["valid"],
        len(dataset["valid"]),
        validation=True,
        wandb_bool=wandb_bool,
        transforms=normalize,
        mode=looper_mode,
    )

    # current best results (lowest mean absolute error on validation set)
    if lr_mode == "min":
        current_best = 100
        second_best = np.infty
    elif lr_mode == "max":
        current_best = 0
        second_best = -1
    for epoch in range(config.epochs):
        # turn off freezing layers after freeze_threshold epochs
        if epoch == freeze_threshold and looper_mode == "classifier":
            train_looper.freeze_layers("off")
            print("Layers were unfrozen")

        print(f"Epoch {epoch + 1}\n")

        # run training epoch and update learning rate
        train_looper.run()

        # run validation epoch
        with torch.no_grad():
            result = valid_looper.run()

        # update learning rate
        lr_scheduler.step(result)

        # update checkpoint if new best is reached
        filename = (
            f"{dataset_name}_"
            f"{network_architecture}_"
            f"{epoch}_"
            f"{result:.4f}.pth"
        )
        if lr_mode == "min":
            if result < current_best:
                current_best = result
                if result < 3:
                    torch.save(
                        network.state_dict(),
                        filename,
                    )

                print(f"\nNew best result: {result}")
            elif result <= second_best:
                second_best = result
                if result < 3:
                    torch.save(
                        network.state_dict(),
                        filename,
                    )
                print(f"\nNew best second result: {result}")
        elif lr_mode == "max":
            if result > current_best:
                current_best = result
                if result >= 0.7:
                    torch.save(
                        network.state_dict(),
                        filename,
                    )

                print(f"\nNew best result: {result}")
            elif result >= second_best:
                second_best = result
                if result >= 0.7:
                    torch.save(
                        network.state_dict(),
                        filename,
                    )
                print(f"\nNew best second result: {result}")
        print("\n", "-" * 80, "\n", sep="")

    print(f"[Training done] Best result: {current_best}")
    torch.save(network.state_dict(), f"{dataset_name}_last.pth")
    if wandb_bool:
        run.finish()


if __name__ == "__main__":
    print("Training parameters were taken from the config file")
    train(**train_params)



__version__ = "0.5.0"
from .data import PolonyDataset, generate_polony_data, make_dataset
from .models import Classifier, UNet, evaluate, predict, predict_one_image, train

__all__ = [
    "PolonyDataset",
    "generate_polony_data",
    "make_dataset",
    "UNet",
    "evaluate",
    "predict",
    "predict_one_image",
    "train",
    "Classifier",
]



"""
This script predicts the number of points in squares on images using
a pre-trained model.

The script takes two arguments:
  1. --path or -p: Path to the folder or file containing the images
    to be processed.
  2. --path_to_model or -m: Path to the saved model state dictionary used
    for predictions.

The prediction results are saved in a file named 'prediction' in a temporary
folder.

Usage example:
python predict_model.py --path /path/to/images --path_to_model /path/to/model

"""

import argparse
import os
import tempfile

from .utils import predict

# Configuring Argument parser
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument(
    "--path",
    "-p",
    type=str,
    default=".",
    help="Path to folder or file with images",
)
parser.add_argument(
    "--path_to_model",
    "-m",
    type=str,
    default="../models/polony_49_1.7496.pth",
    help="Path to saved state dict",
)


def main(args):
    temp_folder = tempfile.mkdtemp()
    print("Creating temporary folder: {}".format(temp_folder))

    predictions = predict(
        path=args.path,
        path_to_model=args.path_to_model,
    )

    with open(os.path.join(temp_folder, "prediction"), "w") as file:
        for pred in predictions:
            for k, v in pred.items():
                file.write(f"Square {k}, number of points {v}\n")


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)



import argparse
import json
import os
import shutil
from glob import glob
from random import random
from typing import Iterable, List, Optional

import h5py
import numpy as np
import yaml
from PIL import Image
from torch.utils.data import Dataset

from .utils import (
    count_data_size,
    create_density_roi,
    delete_duplicates,
    get_and_unzip,
    get_roi_coordinates,
    grid_to_squares,
    read_tiff,
    rgb_to_gray,
)

# folder to load config file
current_script_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.dirname(current_script_path)
CONFIG_PATH = os.path.join(root_path, "config", "config.yaml")

with open(CONFIG_PATH, "r") as file:
    config = yaml.load(file, Loader=yaml.FullLoader)

if isinstance(config["json_path"], List):
    JSON_PATH = "/".join(config["json_path"])
elif isinstance(config["json_path"], str):
    JSON_PATH = config["json_path"]

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument(
    "--train_size",
    "-s",
    type=int,
    default=config["generate_polony_data"]["train_size"],
    help="Percentage of the sample that is used for training",
)

parser.add_argument(
    "--download",
    "-d",
    type=bool,
    default=config["generate_polony_data"]["download"],
    help="Download data or don't",
)

parser.add_argument(
    "--path_to_data",
    "-p",
    type=str,
    default=config["path_to_data"],
    help="Path to dir with data or to dir for download data",
)

parser.add_argument(
    "--id_list_path",
    "-i",
    type=str,
    default=config["id_list_path"],
    help="Path to file with id list for gdown download",
)


class PolonyDataset(Dataset):
    """PyTorch dataset for HDF5 files generated with `get_data.py`."""

    def __init__(
        self,
        dataset_path: List[str] | str,
        horizontal_flip: float,
        vertical_flip: float,
        to_gray: bool,
        json_path: List[str] | str,
        mode: str = "density",
    ):
        """
        Initialize flips probabilities and pointers to a HDF5 file.

        Args:
            dataset_path: a path to a HDF5 file
            horizontal_flip: the probability of applying horizontal flip
            vertical_flip: the probability of applying vertical flip
            mode (str): for which model Dataset - 'density' or 'classifier'
        """
        super(PolonyDataset, self).__init__()
        if isinstance(dataset_path, List):
            self.dataset_path = "/".join(dataset_path)
        elif isinstance(dataset_path, str):
            self.dataset_path = dataset_path
        self.h5 = h5py.File(self.dataset_path, "r")
        self.images = self.h5["images"]
        self.labels = self.h5["labels"]
        self.n_points = self.h5["n_points"]
        self.path_id = self.h5["path"]
        self.square_class = self.h5["class"]
        self.horizontal_flip = horizontal_flip
        self.vertical_flip = vertical_flip
        self.to_gray = to_gray
        if isinstance(json_path, List):
            self.json_path = "/".join(json_path)
        elif isinstance(json_path, str):
            self.json_path = json_path
        with open(self.json_path, "r") as file:
            self.path_dict = json.load(file)
        self.mode = mode

    def __len__(self):
        """Return no. of samples in HDF5 file."""
        return len(self.images)

    def pos_weight(self) -> float:
        """Returns a weight of positive examples for BCELoss

        Returns:
            float: a weight of positive examples for BCELoss
        """
        pos = 0.0
        neg = 0.0
        for i in range(len(self)):
            if self[i][1].item() == 0:
                neg += 1
            else:
                pos += 1
        self.postive_objects = pos
        self.negative_objects = neg
        return pos / neg

    def __getitem__(self, index: int):
        """Return next sample (randomly flipped)."""
        # if both flips probabilities are zero return an image and a label
        if not (self.horizontal_flip or self.vertical_flip):
            image = self.images[index]
            label = self.labels[index]
            n_points = self.n_points[index]
            path = self.path_dict[str(int(self.path_id[index][0]))]
            if self.to_gray:
                image = rgb_to_gray(image)
            if self.mode == "classifier":
                square_class = self.square_class[index]
                return image, square_class, path
            return image, label, n_points, path

        # axis = 1 (vertical flip), axis = 2 (horizontal flip)
        axis_to_flip = []

        if random() < self.vertical_flip:
            axis_to_flip.append(1)

        if random() < self.horizontal_flip:
            axis_to_flip.append(2)
        if self.mode == "classifier":
            return (
                np.flip(self.images[index], axis=axis_to_flip).copy(),
                self.square_class[index],
                self.path_dict[str(int(self.path_id[index][0]))],
            )
        return (
            np.flip(self.images[index], axis=axis_to_flip).copy(),
            np.flip(self.labels[index], axis=axis_to_flip).copy(),
            self.n_points[index],
            self.path_dict[str(int(self.path_id[index][0]))],
        )


def create_empty_hdf5_files(
    dataset_name: str,
    train_size: Optional[int],
    valid_size: int,
    img_size: List[int],
    in_channels: int,
    root_path: str,
):
    """
    Create empty training and validation HDF5 files (one file for training and
    one for validation) with placeholders for images and labels (density maps).

    Note:
    Datasets are saved in [dataset_name]/train.h5 and [dataset_name]/valid.h5.
    Existing files will be overwritten.

    Args:
        dataset_name: used to create a folder for train.h5 and valid.h5
        train_size: no. of training samples, if None -> evaluation mode
        valid_size: no. of validation samples
        img_size: (width, height) of a single image / density map
        in_channels: no. of channels of an input image
        root_path: path to root dir for h5 files

    Returns:
        A tuple of pointers to training and validation HDF5 files.
    """
    # create output folder if it does not exist
    dataset_path = os.path.join(root_path, dataset_name)
    os.makedirs(dataset_path, exist_ok=True)

    # if files exist - delete them
    if os.path.exists(os.path.join(dataset_path, "train.h5")):
        os.remove(os.path.join(dataset_path, "train.h5"))
    if os.path.exists(os.path.join(dataset_path, "valid.h5")):
        os.remove(os.path.join(dataset_path, "valid.h5"))

    # create HDF5 files: [dataset_path]/(train | valid).h5
    if train_size is not None:
        train_h5 = h5py.File(os.path.join(dataset_path, "train.h5"), "w")
    else:
        train_h5 = None
    valid_h5 = h5py.File(os.path.join(dataset_path, "valid.h5"), "w")

    # add two HDF5 datasets (images and labels) for each HDF5 file
    for h5, size in ((train_h5, train_size), (valid_h5, valid_size)):
        if h5 is not None:
            h5.create_dataset("images", (size, in_channels, *img_size))
            h5.create_dataset("path", (size, 1))
            h5.create_dataset("labels", (size, 1, *img_size))
            h5.create_dataset("n_points", (size, 1)),
            h5.create_dataset("class", (size, 1)),

    return train_h5, valid_h5


def generate_polony_data(
    download: bool = config["generate_polony_data"]["download"],
    train_size: int = config["generate_polony_data"]["train_size"],
    new_size: Optional[List[int]] = config["generate_polony_data"]["new_size"],
    data_root: str = config["generate_polony_data"]["data_root"],
    is_squares: bool = config["generate_polony_data"]["is_squares"],
    id_list: Optional[Iterable[str]] = config["generate_polony_data"]["id_list"],
    channels: int = config["generate_polony_data"]["channels"],
    evaluation: bool = config["generate_polony_data"]["evaluation"],
    delete_data: bool = config["generate_polony_data"]["delete_data"],
    is_path: bool = config["generate_polony_data"]["is_path"],
    mode: str = "density",
):
    """
    Generate HDF5 files for polony dataset.

    Args:
        id: zip id on gdrive
        train_size: % from 0 to 100 from data for train
        new_size: - if not None, then the new image size is specified
        download: bool - download data or no
        is_squares: bool - divide into squares or no
        mode: str - for which model ('density' or 'classifier') should be prepared data
    """
    if mode == "classifier":
        is_squares = True
        new_size = [256] * 2
    if not download:
        delete_data = False
    # download and extract dataset
    data_path = os.path.join(data_root, "polony")
    if download:
        for i, id_to_zip in enumerate(id_list):
            location = os.path.join(data_path, str(i))
            get_and_unzip(id_to_zip, location=location)
        delete_duplicates(data_path)

    if new_size is None:
        if is_squares:
            img_size = config["square_size"]
        else:
            img_size = config["img_size"]
    else:
        img_size = new_size

    image_list = []
    for i in range(len(os.listdir(data_path))):
        image_list += glob(os.path.join(data_path, str(i), "slides", "*.tif"))

    names_list = np.array([s.split("/")[-1] for s in image_list])
    sort_idx = names_list.argsort()
    image_list = np.array(image_list)[sort_idx]

    if is_squares:
        # count the number of squares in all images that contain dots
        n_data = count_data_size(image_list, mode=mode)
        if not evaluation:
            train_size = int((train_size / 100) * n_data)
            valid_size = n_data - train_size
            print(f"Data size {n_data}, training size {train_size}")
        else:
            train_size = None
            valid_size = n_data
            print(f"Data size {n_data}")
    else:
        n_data = len(image_list)
        if not evaluation:
            train_size = int((train_size / 100) * n_data)
            valid_size = n_data - train_size
        else:
            train_size = None
            valid_size = n_data

    # create training and validation HDF5 files
    train_h5, valid_h5 = create_empty_hdf5_files(
        dataset_name="polony",
        train_size=train_size,
        valid_size=valid_size,
        img_size=img_size,
        in_channels=channels,
        root_path=data_root,
    )

    # creating a dictionary of paths to collect information in the process of
    # launching fill_h5 function
    path_dict = dict()

    def fill_h5(
        h5,
        images,
        new_size=new_size,
        is_squares=is_squares,
        h5_val=None,
        train_size=train_size,
        valid_size=valid_size,
        channels=channels,
        mode=mode,
    ):
        """
        Save images and labels in given HDF5 file.

        Args:
            h5: HDF5 file
            images: the list of images paths
        """
        if h5 is None:
            h5 = h5_val
            h5_val = None
            train_size = valid_size
        train_j = 0
        val_j = 0
        for i, img_path in enumerate(images):
            # collecting a dictionary of paths {id: path_to_image}
            # (for future analysis of results)
            if i not in path_dict:
                path_dict[i] = img_path
            if is_squares:
                squares_list = grid_to_squares(img_path, new_size=new_size, mode=mode)
                for tt, square_dict in enumerate(squares_list):
                    if channels == 1:
                        image = square_dict["square"]
                    elif channels == 2:
                        image = square_dict["square_2c"]

                    label = square_dict["label"]
                    # get number of points for image
                    n_points = square_dict["n_points"]

                    # for classifier
                    square_class = square_dict["class"]

                    if train_j < train_size:
                        # save data to HDF5 file
                        h5["images"][train_j] = image
                        h5["path"][train_j] = i
                        h5["labels"][train_j, 0] = label
                        h5["n_points"][train_j] = n_points
                        # for classifier
                        h5["class"][train_j] = square_class
                        train_j += 1
                    elif h5_val is not None:
                        h5_val["images"][val_j] = image
                        h5_val["path"][val_j] = i
                        h5_val["labels"][val_j, 0] = label
                        h5_val["n_points"][val_j] = n_points
                        # for classifier
                        h5_val["class"][val_j] = square_class
                        val_j += 1
            else:
                # get an image as numpy array
                if new_size is None:
                    if channels == 1:
                        image = np.array(Image.open(img_path), dtype=np.float32) / 255
                    elif channels == 2:
                        image = read_tiff(img_path)
                else:
                    if channels == 1:
                        image = Image.open(img_path)
                        image = image.resize((new_size[1], new_size[0]))
                        image = np.array(image, dtype=np.float32) / 255
                    elif channels == 2:
                        image = read_tiff(img_path, new_size=new_size)
                # add dim=0 to shape
                if channels == 1:
                    image = np.expand_dims(image, 0)

                # load an RGB image
                coordinates = get_roi_coordinates(img_path, channel=1)
                label = create_density_roi(coordinates, new_size=new_size)

                # get number of points for image
                # TODO make getting the number of points
                n_points = len(coordinates)

                # save data to HDF5 file
                h5["images"][i] = image
                h5["labels"][i, 0] = label
                h5["n_points"][i] = n_points
        print(f"Train size {train_j}, validation size {val_j}")

    if is_squares:
        fill_h5(train_h5, image_list, h5_val=valid_h5)
    else:
        if train_h5 is not None:
            fill_h5(train_h5, image_list[:train_size])
        fill_h5(valid_h5, image_list[train_size:])

    if is_path:
        # writing a path_dict to a json file for further use
        with open(JSON_PATH, "w") as file:
            json.dump(path_dict, file)

    # close HDF5 files
    if train_h5 is not None:
        train_h5.close()
    valid_h5.close()

    if delete_data:
        try:
            for i in range(len(os.listdir(data_path))):
                shutil.rmtree(os.path.join(data_path, str(i)))
        except FileNotFoundError:
            print("File already was deleted")


def main(args):
    id_list = []
    if args.download and os.path.exists(args.id_list_path):
        with open(args.id_list_path, "r") as f:
            for line in f.readlines():
                id_list.append(line.strip())
    elif args.download:
        id_list = config.id_list

    generate_polony_data(
        download=args.download,
        train_size=args.train_size,
        id_list=id_list,
    )


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
